#            ======================================================
#		  library "elliptical.ts" for R  and function "QQplot"
#            ====================================================== 
#
#            Copyright (C) 2016 Vincius Quintas Souto Maior and
#                               Francisco Jos A. Cysneiros
#
#  This library is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published
#  by the Free Software Foundation; either version 2 of the License,
#  or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this library; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  
#  USA or look up the web page http://www.gnu.org/copyleft/gpl.html.
#
#  Please send any comments, suggestions or errors found to:
#  Vincius Quintas Souto Maior, Departamento de Estatstica - CCEN,
#  Universidade Federal de Pernambuco 
#  Recife (PE), Brazil.  Email: viniciussoutomaior@gmail.com 
# ==============================================================================


library(Matrix)

elliptical.ts <- function(Y,X=cbind(seq(1,1,length=length(Y))),family="Normal",epsilon=0.0001,maxit=100,
                 trace="F",np,nq,df=3,epw=0.3,dfs=1,alpha=1,mp=1,epsi=1,sigmap=1,
                 fixed=NULL,...)
{
   n = length(Y)
   m = max(np,nq) 
   p = ncol(X)-1
   
   # Initialize values

   beta = NULL
   phi = matrix(0,nrow=np,ncol=1)   
   theta = matrix(0,nrow=nq,ncol=1)
   aux.model = arima(Y,order=c(np,0,nq),xreg=X,include.mean=FALSE,fixed=fixed) 
   beta <- aux.model$coef[(np+nq+1):(np+nq+1+p)]
   if(np!=0) phi = aux.model$coef[1:np]
   if(nq!=0) theta = aux.model$coef[(1+np):(np+nq)]

   delta <- c(beta,phi,theta)
   dispersao <- aux.model$sigma2
   XB = X*beta
   residuos = aux.model$residuals
   z <- residuos[(m+1):n]/sqrt(dispersao)

   # Initialize arguments

   A = matrix(0,nrow=(n-m),ncol=np)
   B = matrix(0,nrow=(n-m),ncol=nq)
   C = matrix(0,nrow=(n-m),ncol=(p+1))
   aux = matrix(0,nrow=(n-m),ncol=np)
   x = matrix(0,nrow=(n-m),ncol=(p+1))
 
   if(np!=0)  for(i in 1:np)  A[,i] = Y[(m+1-i):(n-i)] - XB[(m+1-i):(n-i)]
   A[,1:ncol(A)]=A[,ncol(A):1]  ##
   if(nq!=0)  for(i in 1:nq)  B[,i] = residuos[(m+1-i):(n-i)]  
   
   for(i in 1:(p+1))  
   {   if(np!=0)
       {  for(j in 1:np)  aux[,j] = X[(m+1-j):(n-j),i]
          aux[,1:ncol(aux)]=aux[,ncol(aux):1]  ##
          C[,i] = X[(m+1):n,i] - aux%*%phi
       }
       if(np==0) C[,i] = X[(m+1):n,i]
   }
   
   if(is.null(fixed)=="FALSE")
   {  if(np!=0 && nq==0)
      {  for(i in 1:np)
         {  if(is.na(fixed[i])=="FALSE") 
              A[,i]=0
         }
      }
      if(np==0 && nq!=0)
      {  for(i in 1:nq)
         {  if(is.na(fixed[i])=="FALSE") 
              B[,i]=0
         }
      }
      if(np!=0 && nq!=0)
      {  for(i in 1:np)
         {  if(is.na(fixed[i])=="FALSE") 
              A[,i]=0
         }
         for(i in (np+1):(np+nq))
         {  if(is.na(fixed[i])=="FALSE") 
              B[,(i-np)]=0
         }
      }
   }

   O = cbind(C,A,B)
   start = delta
   
   if(is.null(fixed)=="FALSE")
   {   O.new = matrix(0,nrow=(n-m),ncol=table(is.na(fixed))["TRUE"])
       j=1
       for(i in 1:ncol(O))
       {  if(sum(O[,i])!=0)
          {  O.new[,j] = O[,i]
             j=j+1
          }
       }
       delta.new = matrix(0,nrow=table(is.na(fixed))["TRUE"],ncol=1)
       j=1
       for(i in 1:length(delta))
       {  if(delta[i]!=0)
          {  delta.new[j] = delta[i]
             j=j+1
          }
       }
   O = O.new
   start = delta.new
   }

   x = X[(m+1):n,]*beta
   MU = x + A%*%phi + B%*%theta
   Y.=Y[(m+1):n]
  
   # Iterative process
 
   iter <- 1
   error1 <- error2 <- 0
   repeat 
   {
      if(family=="Normal")
      { g0 <- log(1/sqrt(2*pi)*exp(-0.5*z^2))
        w.1 <- rep(-0.5,length(MU))
        w.2 <- rep(0,length(z))
        dg <- 1/4
        fg <- 3/4  
        scalevar <- 1
      }
      if(family=="Student")
      { g0 <- log(df^(df/2)*gamma(0.5+df/2)/(gamma(0.5)*gamma(df/2))*(df+z^2)^(-0.5*(df+1)))
        w.1 <- (df+1)/(-2*(df+z^2))
        w.2 <- (df+1)/(2*(df+z^2)^2)
        dg <- (df+1)/(4*(df+3))
        fg <- (3*(df+1))/(4*(df+3))
        scalevar <- df/(df-2)
      }   
      if(family=="Gstudent")
      { g0 <- log((dfs^(df/2)*gamma(0.5+df/2))/(gamma(0.5)*gamma(df/2))*(dfs+z^2)^(-0.5*(df+1)))
        w.1 <- (df+1)/(-2*(dfs+z^2))
        w.2 <- (df+1)/(2*(dfs+z^2)^2)
        dg <- (df*(df+1))/(4*s*(df+3))
        fg <- (3*(df+1))/(4*(df+3))
        scalevar <- dfs/(df-2)
      }
      if(family=="ExpPower")
      { g0 = log(1/(gamma(1+((1+epw)/2))*2^(1+(1+epw)/2))*exp(-0.5*(abs(z)^(2/(1+epw)))))
	  w.1 = 1/(-2*(1+epw)*(z^2)^(epw/(1+epw))) 
        w.2 = epw/(2*(z^2)^((2*epw+1)/(1+epw))*((1+epw)^2))
	  dg = (gamma((3-epw)/2))/(4*(2^(epw-1)*(1+epw)^2*gamma((epw+1)/2)))
	  fg = (epw+3)/(4*(epw+1)) 
        scalevar <-  2^(1+epw)*(gamma(1.5*(epw+1))/(gamma((epw+1)/2)))
      }
      if(family=="LogisticI")
      { g0 = log(1.4843300029*exp(-z^2)/(1+exp(-z^2))^2)
	  w.1 = -tanh(z^2/2)
        w.2 = -0.5+0.5*tanh(0.5*z^2)^2
	  dg = 1.477240176/4
	  fg = 4.013783934/4
        scalevar <- 0.79569
      }
      if(family=="LogisticII")
      { g0 = log(exp(z)/(1+exp(z))^2)
	  w.1 = (exp(z)-1)/(-2*(z*(1+exp(z))))
        w.2 = (2*exp(z)*z-exp(2*z)+1)/(4*z^3*(1+exp(z)^2))
	  dg = 1/12
	  fg = 2.42996/4
        scalevar <- pi^2/3
      }
      if(family=="Glogistic")
      { g0 = log((alpha*gamma(mp+mp))/(gamma(mp)*gamma(mp))*
		   (exp(alpha*z)/(1+exp(alpha*z))^2)^mp)
	  w.1 = alpha*mp*(exp(alpha*z)-1)/(-2*(z*(1+exp(alpha*z))))
        w.2 = alpha*mp*(2*alpha*exp(alpha*z)*z-exp(2*alpha*z)+1)/(4*z^3*(1+exp(alpha*z)^2))
	  dg = (alpha^2*mp^2)/(4*(2*mp+1))
	  fg = (2*m)*(2+mp^2*trigamma(mp))/(4*(2*mp+1))
        scalevar <- 2*trigamma(mp)
      }
      if(family=="ExpPower")
      { g0 = log(1/(gamma(1+((1+epw)/2))*2^(1+(1+epw)/2))*exp(-0.5*(abs(z)^(2/(1+epw)))))
	  w.1 = 1/(-2*(1+epw)*(z^2)^(epw/(1+epw))) 
        w.2 = epw/(2*(z^2)^((2*epw+1)/(1+epw))*((1+epw)^2))
	  dg = (gamma((3-epw)/2))/(4*(2^(epw-1)*(1+epw)^2*gamma((epw+1)/2)))
	  fg = (epw+3)/(4*(epw+1)) 
        scalevar <-  2^(1+epw)*(gamma(1.5*(epw+1))/(gamma((epw+1)/2)))
      }
      if(family=="Cauchy")
      { g0 <- log((1/pi)*(1+z^2)^(-1))
        w.1 <- -1/(1+z^2)
        w.2 <- 1/((1+z^2)^2)
        dg <- 1/8
        fg <- 3/8  
        scalevar <- 1 # non exist
      }
      if(family=="Cnormal")
      { g0 = log((1-epsi)*1/(sqrt(2*pi))*exp(-0.5*z^2)+
		 epsi*1/(sqrt(2*pi)*sigmap)*exp(-0.5*z^2/sigmap^2))
	  w.1 = ((1-epsi)*exp(-z^2/2)+(epsi*(sigmap^2)^(-1.5)*exp(-z^2/(2*sigmap^2))))/
	        ((-2)*((1-epsi)*exp(-z^2/2)+(epsi*(sigmap^2)^(-0.5)*exp(-z^2/(2*sigmap^2)))))
        w.2 = NULL
	  dg = NULL
	  fg = NULL
        scalevar <- 1+epsi*(sigmap^2-1)
      }

      D <- diag(as.vector(-2*w.1))
      z_delta = O%*%start + 1/(4*dg)*D%*%(Y.-MU)       
      new.start <- solve(t(O)%*%O)%*%t(O)%*%z_delta
      
 	error1 <- max(abs((new.start-start)/start))
	start <- new.start

      if(is.null(fixed)=="FALSE")
      {   new.start2 = matrix(0,nrow=(p+1+np+nq),ncol=1)
          new.start2[1] = new.start[1]
          j=2
          for(i in 2:(p+1+np+nq))
          {  if(is.na(fixed[i-1])=="TRUE") 
             {  new.start2[i] = new.start[j]
                j=j+1
             }
          }
          new.start = new.start2
      }   

      beta <- new.start[1:(p+1)]
      if(np!=0) phi <- new.start[(p+2):(1+p+np)]
      if(nq!=0) theta <- new.start[(p+2+np):(1+p+np+nq)]
      XB = X%*%beta
    
      # New residuals
   
      if(np!=0 && nq==0)
      {  residuos = NULL  # original scale of measurement   
         new.A = matrix(0,nrow=(n-m),ncol=np)
         for(i in 1:np)
            new.A[1:(n-np),i] = Y[(np+1-i):(n-m+np-i)] - X[(np+1-i):(n-m+np-i)]*beta
         new.A[,1:ncol(new.A)]=new.A[,ncol(new.A):1]  ##
         residuos = Y[(m+1):n] - X[(m+1):n,]*beta - new.A%*%phi
      }
      if(np==0 && nq!=0)
      {  residuos = NULL  # original scale of measurement 
         new.B = matrix(0,nrow=n,ncol=nq)
         for(i in 1:n)
         {  residuos[i] = Y[i] - X[i,]*beta - new.B[i,]%*%theta
            for(j in 1:nq)
                if(j<=i && i<n)  new.B[(i+1),j] = residuos[i+1-j]
         }     
      }
      if(np!=0 && nq!=0)
      {  residuos = NULL  # original scale of measurement 
         new.A = matrix(0,nrow=n,ncol=np)
         new.B = matrix(0,nrow=n,ncol=nq)
         for(i in 1:n)
         {  residuos[i] = Y[i] - X[i,]*beta - new.A[i,]%*%phi - new.B[i,]%*%theta
            for(j in 1:np)
                if(j<=i && i<n)  new.A[(i+1),j] = Y[i+1-j] - X[i+1-j]*beta
            for(j in 1:nq)
                if(j<=i && i<n)  new.B[(i+1),j] = residuos[i+1-j]
         }
      }
      if(np==0 && nq==0)
      {  residuos = NULL  # original scale of measurement 
         residuos = Y[1:n] - X[1:n,]*beta
      }

      # New arguments
 
      if(np!=0)  for(i in 1:np)  A[,i] = Y[(m+1-i):(n-i)] - XB[(m+1-i):(n-i)]
      A[,1:ncol(A)]=A[,ncol(A):1] ##
      if(nq!=0)  for(i in 1:nq)  B[,i] = residuos[(m+1-i):(n-i)]  

      for(i in 1:(p+1))  
      {   if(np!=0)
          {   for(j in 1:np)  aux[,j] = X[(m+1-j):(n-j),i]
              aux[,1:ncol(aux)]=aux[,ncol(aux):1] ##
              C[,i] = X[(m+1):n,i] - aux%*%phi
          }
          if(np==0)  C[,i] = X[(m+1):n,i]
      }
   

      if(is.null(fixed)=="FALSE")
      {  if(np!=0 && nq==0)
         {  for(i in 1:np)
            {  if(is.na(fixed[i])=="FALSE") 
                 A[,i]=0
            }
         }
         if(np==0 && nq!=0)
         {  for(i in 1:nq)
            {  if(is.na(fixed[i])=="FALSE") 
                 B[,i]=0
            }
         }
         if(np!=0 && nq!=0)
         {  for(i in 1:np)
            {  if(is.na(fixed[i])=="FALSE") 
                 A[,i]=0
            }
            for(i in (np+1):(np+nq))
            {  if(is.na(fixed[i])=="FALSE") 
                 B[,(i-np)]=0
            }
         }
      }

      O = cbind(C,A,B)
   
      if(is.null(fixed)=="FALSE")
      {   O.new = matrix(0,nrow=(n-m),ncol=table(is.na(fixed))["TRUE"])
          j=1
          for(i in 1:ncol(O))
          {  if(sum(O[,i])!=0)
             {  O.new[,j] = O[,i]
                j=j+1
             }
          }
          O = O.new
      }

      x = X[(m+1):n,]*beta
      MU = x + A%*%phi + B%*%theta
      Y.=Y[(m+1):n]
     
      new.dispersao <- 1/(n-m)*t(Y.-MU)%*%D%*%(Y.-MU)

      error2 <- abs((new.dispersao-dispersao)/dispersao)
      dispersao <- c(new.dispersao)
      if(np!=0 && nq==0)  z <- residuos/sqrt(dispersao)
      if(np!=0 && nq!=0)  z <- residuos[(m+1):n]/sqrt(dispersao)
      if(np==0 && nq!=0)  z <- residuos[(m+1):n]/sqrt(dispersao)
      if(np==0 && nq==0)  z <- residuos[(m+1):n]/sqrt(dispersao)
      
      if((iter == maxit) || (max(error1, error2, na.rm = TRUE) < epsilon))
 	     break

      iter <- iter + 1
   }

   if(maxit>1 && iter==maxit)
   {
      warning(paste("\n linear convergence not obtained in", maxit,
                    "iterations"))
   }
     
   desviopadrao <- sqrt(diag(solve((4*dg/c(dispersao))*t(O)%*%O)))
   if(is.null(fixed)=="FALSE")
   {   desvio2 = matrix(0,nrow=(p+1+np+nq),ncol=1)
       desvio2[1] = desviopadrao[1]
       j=2
       for(i in 2:(p+1+np+nq))
       {  if(is.na(fixed[i-1])=="TRUE") 
          {  desvio2[i] = desviopadrao[j]
             j=j+1
          }
       }
       desviopadrao = desvio2
   }   
   desviopadrao2 <- sqrt(1/((n-m)*(4*fg-1)/(4*c(dispersao)^2)))
   loglik <- -0.5*length(MU)*log((dispersao)) + sum(g0)
   AIC <- -2*loglik + 2*(np+nq+1+1)
   BIC <- -2*loglik + (np+nq+1+1)*log(length(MU))
   RMSE <- sqrt(sum(((Y. - MU)/Y.)^2)/length(Y.))

   if(trace == "TRUE")
   {  
	print(matrix(nrow=2, ncol=0, byrow=TRUE,dimnames = list(c("Call:", paste(c("symarma(",np,",",nq,") - family: ",family),collapse="")))))    
	print(matrix(nrow=1, ncol=0, byrow=TRUE,dimnames = list(c("Coefficients:"))))
	if(np!=0 && nq!=0)
	   print(matrix(c(new.start,desviopadrao), nrow=np+nq+1, ncol=2, byrow=FALSE, dimnames = list(c("intercept",paste("ar", 1:np, sep = ""),paste("ma", 1:nq, sep = "")), c("Estimative","s.d."))))
	if(np!=0 && nq==0)
	   print(matrix(c(new.start,desviopadrao), nrow=np+nq+1, ncol=2, byrow=FALSE, dimnames = list(c("intercept",paste("ar", 1:np, sep = "")), c("Estimative","s.d."))))
	if(np==0 && nq!=0)
	   print(matrix(c(new.start,desviopadrao), nrow=np+nq+1, ncol=2, byrow=FALSE, dimnames = list(c("intercept",paste("ma", 1:nq, sep = "")), c("Estimative","s.d."))))
	print(matrix(nrow=3, ncol=0, byrow=TRUE,dimnames =list(c(
        paste(c("varphi estimated as ",sprintf("%0.4f", dispersao),"  (s.d. = ",sprintf("%0.4f",desviopadrao2),")"),collapse=""),
        paste(c("log likelihooh = ",sprintf("%0.4f", loglik)),collapse=""),
        paste(c("rmse = ",sprintf("%0.2f", RMSE)),collapse="")))))
   }

   fit <- list(coefficients=new.start, dispersion=dispersao,residuals=(Y.-MU)/sqrt(c(dispersao)),
               fitted.values=MU,loglik=loglik,Wg=w.1,Wgder=w.2,iter=iter,fun.g=g0,erro=(Y.-MU),
               scale=(4*dg),scaledispersion=(4*fg-1),scalevariance=scalevar,A=A,B=B,C=C,O=O,n=length(Y),
               DesP.C=desviopadrao,DesP.D=desviopadrao2,rmse=RMSE,family=family,df=df,np=np,nq=nq,
               epw=epw,family=family,fixed=fixed,X=X)
}



QQplot<- function(fit, B=300)
{
   resid = fit$residuals
   n = length(resid)
   np = fit$np
   nq = fit$nq
   m = max(np,nq)
   df = fit$df
   mu = fit$fitted.values
   varphi = fit$dispersion
   dist = fit$family
   fixed = fit$fixed
   X = fit$X

   resp<-NULL
   e <- matrix(0,(n-m),B)
   for(i in 1:B)
   {   if(dist=="Normal")
       {  resp <- rnorm(n,0,1)
          resp <- mu + sqrt(varphi)*resp
          fit.s <- elliptical.ts(resp,cbind(X[(m+1):(n+m)]),family=dist,fixed=fixed,np=np,nq=nq,trace=FALSE)
       }
       if(dist=="Student")
       {  resp <- rt(n,df=df)
          resp <- mu + sqrt(varphi)*resp
          fit.s <- elliptical.ts(resp,cbind(X[(m+1):(n+m)]),family=dist,fixed=fixed,df=df,np=np,nq=nq,trace=FALSE)
       }  
       td <- fit.s$residuals
       e[,i] <- sort(td)
       e1 <- numeric(n-m)
       e2 <- numeric(n-m)
       for(i in 1:(n-m))
       {   eo <- sort(e[i,])
           e1[i] <- eo[ceiling(B*0.05)]
           e2[i] <- eo[ceiling(B*0.95)]
       }
       med <- apply(e,1,mean)
       faixa <- range(resid,e1,e2)
       par(pty="s")
       if(dist=="Normal")
          qqnorm(resid,main = expression("Q-Q plot for Normal"),xlab="Theorical Quantiles",ylab="Residuals Standardized", ylim=faixa, pch=16)
       if(dist=="Student")
          qqnorm(resid,main = expression("Q-Q plot for" ~~ {t-Student}),xlab="Theorical Quantiles",ylab="Residuals Standardized", ylim=faixa, pch=16)
       par(new=TRUE)
       qqnorm(e1,axes=F,main="",xlab="",ylab="",type="l",ylim=faixa,lty=1)
       par(new=TRUE)
       qqnorm(e2,axes=F,main="",xlab="",ylab="", type="l",ylim=faixa,lty=1)
       par(new=TRUE)
       qqnorm(med,axes=F,main="",xlab="", ylab="", type="l",ylim=faixa,lty=2)
    }
}

