#            ======================================================
#		                function "influence.ts" 
#            ====================================================== 
#
#            Copyright (C) 2016 Vincius Quintas Souto Maior and
#                               Francisco Jos A. Cysneiros
#
#  This library is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published
#  by the Free Software Foundation
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this library; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  
#  USA or look up the web page http://www.gnu.org/copyleft/gpl.html.
#
#  Please send any comments, suggestions or errors found to:
#  Vincius Quintas Souto Maior, Departamento de Estatstica - CCEN,
#  Universidade Federal de Pernambuco 
#  Recife (PE), Brazil.  Email: viniciussoutomaior@gmail.com 
# ==============================================================================
#        "influence.ts" set of functions used by 'elliptical.ts'
# ==============================================================================

library(Matrix)

influence.ts <- function(ajust.model,esq.pert="aditive",seed=2015,benchmarks.rep=2000,
                         perc.ind1=95,perc.ind2=5,diag="incl",plot="TRUE",trace="TRUE")
{   
   n = ajust.model$n
   X = ajust.model$X
   X = cbind(X)
   p = ncol(X)-1

   g0.model = ajust.model$fun.g
   Wg.model = ajust.model$Wg
   Wgder.model = ajust.model$Wgder
   var.model = as.vector(ajust.model$dispersion)
   hat = ajust.model$coefficients
   v.model <- -2*Wg.model
   erro.model = ajust.model$erro
   uwt.model = ajust.model$residuals^2
   A.model <- P <- ajust.model$A
   B.model <- R <- ajust.model$B
   dist<- ajust.model$family
   np <- ajust.model$np
   nq <- ajust.model$nq
   m <- max(np,nq)
   df <- ajust.model$df
   fixed <- ajust.model$fixed
   C.model <- ajust.model$C
   O <- cbind(C.model,A.model,B.model)
   
   ar=ma=NULL
   if(np!=0)  ar[1:np] = hat[2:(np+1)]
   if(np!=0 && nq!=0) ma[1:nq] = hat[(np+2):(np+nq+1)]
   if(np==0 && nq!=0) ma[1:nq] = hat[2:(nq+1)]
 
   if(diag=="incl")
   {
      if(esq.pert=="aditive")
      {  S.exe = -(1/var.model)*v.model*erro.model
	   O.exe = sqrt(sum(S.exe^2))
	   D.exe = 2*S.exe
      }
      if(esq.pert=="dispersion")
      {  S.exe = -0.5 + 0.5*v.model*uwt.model
	   O.exe = sqrt(sum(S.exe^2))
	   D.exe = 2*S.exe
      }

      # Benchmarks

      set.seed(seed)
      rep = benchmarks.rep
      serie = matrix(0,nrow = n, ncol = rep)
      S = matrix(0,nrow = (n-m), ncol = rep)
 	
      loop=1 
      repeat
      {
         for(j in loop:rep)
         {  if(dist=="Normal")
               serie[,j] = arima.sim(n=n,list(ar=ar,ma=ma),sd=sqrt(var.model))
            if(dist=="Student") 
               serie[,j] = arima.sim(n=n,list(ar=ar,ma=ma),rand.gen=function(n, ...) var.model*(df-2)/df*rt(n=n,df=df))
          
            fit <- elliptical.ts(serie[,j],X,family=dist,fixed=fixed,np=np,nq=nq,df=df,trace=FALSE)
            Wg = fit$Wg
            erro = fit$erro
            varphi.hat = as.vector(fit$dispersion)
            g0 = fit$fun.g
            v.loop = -2*Wg
            uwt = fit$residuals^2
            A.loop = fit$A
            B.loop = fit$B

            if(esq.pert=="aditive") S[,j] = -(1/varphi.hat)*v.loop*erro
            if(esq.pert=="dispersion") S[,j] = -0.5 + 0.5*v.loop*uwt  

            if(j==1) print("Benchmarks - 00% ...")   
            if(j==floor(1*rep/4)) print("Benchmarks - 25% ...")
            if(j==floor(2*rep/4)) print("Benchmarks - 50% ...")
            if(j==floor(2*rep/4)) print("Benchmarks - 75% ...")
            if(j==floor(4*rep/4)) print("Benchmarks - 100%")
        }
        loop = j
        if(loop==rep)  break
      }
	
      # "BS0" statistic

      O = NULL
      for(i in 1:rep)    O[i] = sqrt(sum(S[,i]^2))
      BS0 = quantile(O,  probs = c(perc.ind1)/100)

      # "BS1" statistic

      D = matrix(nrow = (n-m), ncol = rep)
      for(i in 1:rep)    D[,i] = 2*S[,i]
      qnt = NULL
      for(i in 1:rep) qnt[i] = max(abs(D[,i]))
      BS1 = quantile(qnt,  probs = c(perc.ind1)/100)

      # "BS2" statistic

      bs2 = NULL
      j=1
      for(i in 1:rep)
      {  if(O[i] > BS0)
         {  bs2[j] = i
            j = j+1
         }
      }
      help = NULL
      for(i in 1:length(bs2))  
      {   k = bs2[i]
          help[i] = max(abs(D[,k]))
      }
      BS2 = quantile(help,  probs = c(perc.ind2)/100)
 
      # Plot
 
      if(plot=="TRUE")
      {  dat = abs(D.exe)[,1]
         barplot(dat,space=1,xlab="Index",ylab="",main=substitute(paste("Billor and Loynes")))
         abline(h=BS1,lty=1)
         abline(h=BS2,lty=2)
      }

      # Print 

      if(trace == "TRUE")
      {  
   	   print(matrix(nrow=2, ncol=0, byrow=TRUE,dimnames = list(c("Call:", paste(c("symarma(",np,",",nq,") - family: ",dist),collapse="")))))    
         print(matrix(nrow=1, ncol=0, byrow=TRUE,dimnames = list(c("Billor & Loynes"))))
         print(matrix(nrow=1, ncol=0, byrow=TRUE,dimnames = list(c("Measures of local influence:"))))
         print(matrix(c(O.exe),nrow=1, ncol=1,byrow=TRUE,dimnames = list(c("Inclination"))))
         print(matrix(nrow=1, ncol=0, byrow=TRUE,dimnames = list(c("Benchmarks:"))))
         print(matrix(c(BS0,BS1,BS2),nrow=3, ncol=1,byrow=FALSE,dimnames = list(c(
            "BS0","BS1","BS2"))))
      }
   }

   if(diag=="cook")
   {
      f1 <- 0.5 + 2*Wg.model*uwt.model + Wgder.model*uwt.model^2
      A <- as.vector(-2*(Wg.model + 2*Wgder.model*uwt.model))
      B <- as.vector((Wg.model + Wgder.model*uwt.model)*erro.model)
      C <- as.vector(Wg.model + Wgder.model*uwt.model)

      # Observed Fisher information matrix

      I11 <- -(1/var.model)*sum(A)     # varphi
      I12 <-  (2/var.model^2)*sum(B)   # beta
      I13 <- -(1/var.model)*t(P)%*%A   # phi
      I14 <- -(1/var.model)*t(R)%*%A   # theta
      I22 <-  (1/var.model^2)*sum(f1)
      I23 <-  (2/var.model^2)*t(P)%*%B
      I24 <-  (2/var.model^2)*t(R)%*%B
      I33 <- -(1/var.model)*t(P)%*%diag(A)%*%P
      I34 <- -(1/var.model)*t(P)%*%diag(A)%*%R
      I44 <- -(1/var.model)*t(R)%*%diag(A)%*%R

      I <- matrix(nrow=(np+nq+p+2), ncol=(np+nq+p+2))
      M1 <-cbind(t(I11),I12,t(I13),t(I14))  
      M2 <-cbind(I12,I22,t(I23),t(I24))
      if(np!=0) M3 <-cbind(I13,I23,I33,I34)
      if(nq!=0) M4 <-cbind(I14,I24,I34,I44)
      I[1,]<- M1
      I[2,]<- M2
      if(np!=0) I[3:(np+2),]<- M3
      if(np!=0 && nq!=0) I[(np+3):(np+nq+2),]<- M4
      if(nq!=0) I[3:(nq+3),]<- M4

      # Delta matrix

      D.aux1=D.aux2=NULL
      if(esq.pert=="aditive") 
      {   D.aux1 <- -(2/var.model^2)*B         # varphi
          D.aux2 <- (1/var.model)*diag(A)%*%O  # delta
          Delta <- matrix(c(D.aux1,D.aux2),ncol=(n-m),byrow=TRUE)
      }
      if(esq.pert=="dispersion")
      {   D.aux1 <- (1/var.model)*diag(C)%*%uwt.model  # varphi
          D.aux2 <- (2/var.model)*diag(B)%*%O          # delta
          Delta <- matrix(c(D.aux1,D.aux2),ncol=(n-m),byrow=TRUE)
      }

      if(is.null(fixed)=="FALSE")
      {   Delta.new = matrix(0,ncol=(n-m),nrow=(1+table(is.na(fixed))["TRUE"]))
          Delta.new[1,] = Delta[1,]
          j=2
          for(i in 2:nrow(Delta))
          {  if(sum(Delta[i,])!=0)
             {  Delta.new[j,] = Delta[i,]
                j=j+1
             }
          }
          I.new = matrix(0,nrow=(1+table(is.na(fixed))["TRUE"]),ncol=(np+nq+p+2))
          I.new[1,] = I[1,]
          j=2
          for(i in 2:nrow(I))
          {  if(sum(I[i,])!=0)
             {  I.new[j,] = I[i,]
                j=j+1
             }
          }
          I.new.new = matrix(0,nrow=(1+table(is.na(fixed))["TRUE"]),ncol=(1+table(is.na(fixed))["TRUE"]))
          I.new.new[,1] = I.new[,1]
          j=2
          for(i in 2:ncol(I.new))
          {  if(sum(I.new[,i])!=0)
             {  I.new.new[,j] = I.new[,i]
                j=j+1
             }
          }
          Delta = Delta.new
          I = I.new.new
      }

      Curvatura <- -2*t(Delta)%*%solve(I)%*%Delta
      OC.exe <- max(abs(eigen(Curvatura)$values))
      C.exe <- eigen(Curvatura)$vectors[,abs(eigen(Curvatura)$values)==max(abs(eigen(Curvatura)$values))]

      # Benchmarks

      set.seed(seed)
      rep = benchmarks.rep
      serie = matrix(0,nrow = n, ncol = rep)
      
      Oc.sim = NULL
      C.sim = matrix(nrow = (n-m), ncol = rep)

      loop=1 
      repeat
      {
         for(j in loop:rep)
         {  if(dist=="Normal")
               serie[,j] = arima.sim(n=n,list(ar=ar,ma=ma),sd=sqrt(var.model))
            if(dist=="Student") 
               serie[,j] = arima.sim(n=n,list(ar=ar,ma=ma),rand.gen=function(n, ...) var.model*(df-2)/df*rt(n=n,df=df))
          
            fit <- elliptical.ts(serie[,j],X,family=dist,fixed=fixed,np=np,nq=nq,df=df,trace=FALSE)
            
            varphi.hat = as.vector(fit$dispersion)
            Wg = fit$Wg
            Wgder = fit$Wgder
            uwt = fit$residuals^2
            erro = fit$erro
            A.fit <- P.fit <- fit$A
            B.fit <- R.fit <- fit$B 
            C.fit <- fit$C
            f1.sim <- 0.5 + 2*Wg*uwt + Wgder*uwt^2
            A.sim <- as.vector(-2*(Wg + 2*Wgder*uwt))
            B.sim <- as.vector((Wg + Wgder*uwt)*erro)
            C.sim.sim <- as.vector(Wg + Wgder*uwt)
            g0 = fit$fun.g
            v = -2*Wg
            O.sim <- cbind(C.fit,A.fit,B.fit)
            
            I11.sim <- -(1/varphi.hat)*sum(A.sim) 
            I12.sim <-  (2/varphi.hat^2)*sum(B.sim) 
            I13.sim <- -(1/varphi.hat)*t(P.fit)%*%A.sim 
            I14.sim <- -(1/varphi.hat)*t(R.fit)%*%A.sim 
            I22.sim <-  (1/varphi.hat^2)*sum(f1.sim)
            I23.sim <-  (2/varphi.hat^2)*t(P.fit)%*%B.sim 
            I24.sim <-  (2/varphi.hat^2)*t(R.fit)%*%B.sim 
            I33.sim <- -(1/varphi.hat)*t(P.fit)%*%diag(A.sim)%*%P.fit
            I34.sim <- -(1/varphi.hat)*t(P.fit)%*%diag(A.sim)%*%R.fit
    
            I.sim <- matrix(nrow=(np+nq+p+2), ncol=(np+nq+p+2))
            M1 <-cbind(t(I11),I12,t(I13),t(I14))  
            M2 <-cbind(I12,I22,t(I23),t(I24))
            if(np!=0) M3 <-cbind(I13,I23,I33,I34)
            if(nq!=0) M4 <-cbind(I14,I24,I34,I44)
            I.sim[1,]<- M1
            I.sim[2,]<- M2
            if(np!=0) I.sim[3:(np+2),]<- M3
            if(np!=0 && nq!=0) I.sim[(np+3):(np+nq+2),]<- M4
            if(nq!=0) I.sim[3:(nq+3),]<- M4

            D.sim.aux1=D.sim.aux2=NULL
            if(esq.pert=="aditive") 
            {   D.sim.aux1 <- -(2/varphi.hat^2)*B.sim             # varphi
                D.sim.aux2 <- (1/varphi.hat)*diag(A.sim)%*%O.sim  # delta
                Delta.sim <- matrix(c(D.sim.aux1,D.sim.aux2),ncol=(n-m),byrow=TRUE)
            }
            if(esq.pert=="dispersion")
            {   D.sim.aux1 <- (1/varphi.hat)*diag(C.sim.sim)%*%uwt       # varphi
                D.sim.aux2 <- (2/varphi.hat)*diag(B.sim)%*%O.sim          # delta
                Delta.sim <- matrix(c(D.sim.aux1,D.sim.aux2),ncol=(n-m),byrow=TRUE)
            }

            if(is.null(fixed)=="FALSE")
            {   Delta.sim.new = matrix(0,ncol=(n-m),nrow=(1+table(is.na(fixed))["TRUE"]))
                Delta.sim.new[1,] = Delta.sim[1,]
                l=2
                for(i in 2:nrow(Delta.sim))
                {  if(sum(Delta.sim[i,])!=0)
                   {  Delta.sim.new[l,] = Delta.sim[i,]
                      l=l+1
                   }
                }
                I.sim.new = matrix(0,nrow=(1+table(is.na(fixed))["TRUE"]),ncol=(np+nq+p+2))
                I.sim.new[1,] = I.sim[1,]
                l=2
                for(i in 2:nrow(I.sim))
                {  if(sum(I.sim[i,])!=0)
                   {  I.sim.new[l,] = I.sim[i,]
                      l=l+1
                   }
                }
                I.sim.new.new = matrix(0,nrow=(1+table(is.na(fixed))["TRUE"]),ncol=(1+table(is.na(fixed))["TRUE"]))
                I.sim.new.new[,1] = I.sim.new[,1]
                l=2
                for(i in 2:ncol(I.sim.new))
                {  if(sum(I.sim.new[,i])!=0)
                   {  I.sim.new.new[,l] = I.sim.new[,i]
                      l=l+1
                   }
                }
                Delta.sim = Delta.sim.new
                I.sim = I.sim.new.new
            }

            Curvatura.sim <- -2*t(Delta.sim)%*%solve(I.sim)%*%Delta.sim
            Oc.sim[j] <- max(abs(eigen(Curvatura.sim)$values))
            C.sim[,j] <- eigen(Curvatura.sim)$vectors[,abs(eigen(Curvatura.sim)$values)==max(abs(eigen(Curvatura.sim)$values))]

            if(j==1) print("Benchmarks - 00% ...")   
            if(j==floor(1*rep/4)) print("Benchmarks - 25% ...")
            if(j==floor(2*rep/4)) print("Benchmarks - 50% ...")
            if(j==floor(2*rep/4)) print("Benchmarks - 75% ...")
            if(j==floor(4*rep/4)) print("Benchmarks - 100%")
        }         
        loop = j
        if(loop==rep)  break
      }

      # "BC0" statistic

      BC0 = quantile(Oc.sim,  probs = c(perc.ind1)/100)

      # "BC1" statistic

      K = matrix(nrow = (n-m), ncol = rep)
      for(i in 1:rep)    K[,i] = C.sim[,i]
      qnt.sim = NULL
      for(i in 1:rep)   qnt.sim[i] = max(abs(K[,i]))
      BC1 = quantile(qnt.sim, probs = c(perc.ind1)/100)
 
      # "BC2" statistic

      bs2 = NULL
      j=1
      for(i in 1:rep)
      {   if(Oc.sim[i] > BC0)
          {  bs2[j] = i
             j = j+1
          }
      }
      help = NULL
      for(i in 1:length(bs2))  
      {   kk = bs2[i]
          help[i] = max(abs(K[,kk]))
      }
      BC2 = quantile(help,  probs = c(perc.ind2)/100)

      # Plot
 
      if(plot=="TRUE")
      {  dat = abs(C.exe)
         barplot(dat,space=1,xlab="Index",ylab="",main=substitute(paste("Cook")))
         abline(h=BC1,lty=1)
         abline(h=BC2,lty=2)
      }

      # Print 

      if(trace == "TRUE")
      {  
   	   print(matrix(nrow=2, ncol=0, byrow=TRUE,dimnames = list(c("Call:", paste(c("symarma(",np,",",nq,") - family: ",dist),collapse="")))))    
         print(matrix(nrow=1, ncol=0, byrow=TRUE,dimnames = list(c("Cook"))))
         print(matrix(nrow=1, ncol=0, byrow=TRUE,dimnames = list(c("Measures of local influence:"))))
         print(matrix(c(OC.exe),nrow=1, ncol=1,byrow=TRUE,dimnames = list(c("Curvature"))))
         print(matrix(nrow=1, ncol=0, byrow=TRUE,dimnames = list(c("Benchmarks:"))))
         print(matrix(c(BC0,BC1,BC2),nrow=3, ncol=1,byrow=FALSE,dimnames = list(c(
            "BC0","BC1","BC2"))))
      }
   }

   if(diag=="lv")
   {
      f1 <- 0.5 + 2*Wg.model*uwt.model + Wgder.model*uwt.model^2
      A <- as.vector(-2*(Wg.model + 2*Wgder.model*uwt.model))
      B <- as.vector((Wg.model + Wgder.model*uwt.model)*erro.model)
      C <- as.vector(Wg.model + Wgder.model*uwt.model)

      # Observed Fisher information matrix

      I11 <- -(1/var.model)*sum(A)     # varphi
      I12 <-  (2/var.model^2)*sum(B)   # beta
      I13 <- -(1/var.model)*t(P)%*%A   # phi
      I14 <- -(1/var.model)*t(R)%*%A   # theta
      I22 <-  (1/var.model^2)*sum(f1)
      I23 <-  (2/var.model^2)*t(P)%*%B
      I24 <-  (2/var.model^2)*t(R)%*%B
      I33 <- -(1/var.model)*t(P)%*%diag(A)%*%P
      I34 <- -(1/var.model)*t(P)%*%diag(A)%*%R
      I44 <- -(1/var.model)*t(R)%*%diag(A)%*%R

      I <- matrix(nrow=(np+nq+p+2), ncol=(np+nq+p+2))
      M1 <-cbind(t(I11),I12,t(I13),t(I14))  
      M2 <-cbind(I12,I22,t(I23),t(I24))
      if(np!=0) M3 <-cbind(I13,I23,I33,I34)
      if(nq!=0) M4 <-cbind(I14,I24,I34,I44)
      I[1,]<- M1
      I[2,]<- M2
      if(np!=0) I[3:(np+2),]<- M3
      if(np!=0 && nq!=0) I[(np+3):(np+nq+2),]<- M4
      if(nq!=0) I[3:(nq+3),]<- M4

      # Delta matrix

      D.aux1=D.aux2=NULL
      if(esq.pert=="aditive") 
      {   D.aux1 <- -(2/var.model^2)*B         # varphi
          D.aux2 <- (1/var.model)*diag(A)%*%O  # delta
          Delta <- matrix(c(D.aux1,D.aux2),ncol=(n-m),byrow=TRUE)
      }
      if(esq.pert=="dispersion")
      {   D.aux1 <- (1/var.model)*diag(C)%*%uwt.model  # varphi
          D.aux2 <- (2/var.model)*diag(B)%*%O          # delta
          Delta <- matrix(c(D.aux1,D.aux2),ncol=(n-m),byrow=TRUE)
      }

      if(is.null(fixed)=="FALSE")
      {   Delta.new = matrix(0,ncol=(n-m),nrow=(1+table(is.na(fixed))["TRUE"]))
          Delta.new[1,] = Delta[1,]
          j=2
          for(i in 2:nrow(Delta))
          {  if(sum(Delta[i,])!=0)
             {  Delta.new[j,] = Delta[i,]
                j=j+1
             }
          }
          I.new = matrix(0,nrow=(1+table(is.na(fixed))["TRUE"]),ncol=(np+nq+p+2))
          I.new[1,] = I[1,]
          j=2
          for(i in 2:nrow(I))
          {  if(sum(I[i,])!=0)
             {  I.new[j,] = I[i,]
                j=j+1
             }
          }
          I.new.new = matrix(0,nrow=(1+table(is.na(fixed))["TRUE"]),ncol=(1+table(is.na(fixed))["TRUE"]))
          I.new.new[,1] = I.new[,1]
          j=2
          for(i in 2:ncol(I.new))
          {  if(sum(I.new[,i])!=0)
             {  I.new.new[,j] = I.new[,i]
                j=j+1
             }
          }
          Delta = Delta.new
          I = I.new.new
      }

      Curvatura <- -2*t(Delta)%*%solve(I)%*%Delta
      OC.exe <- max(abs(eigen(Curvatura)$values))
      Curvatura.lv <- diag(Curvatura)

      # Benchmarks

      set.seed(seed)
      rep = benchmarks.rep
      serie = matrix(0,nrow = n, ncol = rep)
      
      Oc.sim = NULL
      Curvatura.sim.lv = matrix(nrow = (n-m), ncol = rep)

      loop=1 
      repeat
      {
         for(j in loop:rep)
         {  if(dist=="Normal")
               serie[,j] = arima.sim(n=n,list(ar=ar,ma=ma),sd=sqrt(var.model))
            if(dist=="Student") 
               serie[,j] = arima.sim(n=n,list(ar=ar,ma=ma),rand.gen=function(n, ...) var.model*(df-2)/df*rt(n=n,df=df))
          
            fit <- elliptical.ts(serie[,j],X,family=dist,fixed=fixed,np=np,nq=nq,df=df,trace=FALSE)
            
            varphi.hat = as.vector(fit$dispersion)
            Wg = fit$Wg
            Wgder = fit$Wgder
            uwt = fit$residuals^2
            erro = fit$erro
            A.fit <- P.fit <- fit$A
            B.fit <- R.fit <- fit$B 
            C.fit <- fit$C
            f1.sim <- 0.5 + 2*Wg*uwt + Wgder*uwt^2
            A.sim <- as.vector(-2*(Wg + 2*Wgder*uwt))
            B.sim <- as.vector((Wg + Wgder*uwt)*erro)
            C.sim.sim <- as.vector(Wg + Wgder*uwt)
            g0 = fit$fun.g
            v = -2*Wg
            O.sim <- cbind(C.fit,A.fit,B.fit)
            
            I11.sim <- -(1/varphi.hat)*sum(A.sim) 
            I12.sim <-  (2/varphi.hat^2)*sum(B.sim) 
            I13.sim <- -(1/varphi.hat)*t(P.fit)%*%A.sim 
            I14.sim <- -(1/varphi.hat)*t(R.fit)%*%A.sim 
            I22.sim <-  (1/varphi.hat^2)*sum(f1.sim)
            I23.sim <-  (2/varphi.hat^2)*t(P.fit)%*%B.sim 
            I24.sim <-  (2/varphi.hat^2)*t(R.fit)%*%B.sim 
            I33.sim <- -(1/varphi.hat)*t(P.fit)%*%diag(A.sim)%*%P.fit
            I34.sim <- -(1/varphi.hat)*t(P.fit)%*%diag(A.sim)%*%R.fit
    
            I.sim <- matrix(nrow=(np+nq+p+2), ncol=(np+nq+p+2))
            M1 <-cbind(t(I11),I12,t(I13),t(I14))  
            M2 <-cbind(I12,I22,t(I23),t(I24))
            if(np!=0) M3 <-cbind(I13,I23,I33,I34)
            if(nq!=0) M4 <-cbind(I14,I24,I34,I44)
            I.sim[1,]<- M1
            I.sim[2,]<- M2
            if(np!=0) I.sim[3:(np+2),]<- M3
            if(np!=0 && nq!=0) I.sim[(np+3):(np+nq+2),]<- M4
            if(nq!=0) I.sim[3:(nq+3),]<- M4

            D.sim.aux1=D.sim.aux2=NULL
            if(esq.pert=="aditive") 
            {   D.sim.aux1 <- -(2/varphi.hat^2)*B.sim             # varphi
                D.sim.aux2 <- (1/varphi.hat)*diag(A.sim)%*%O.sim  # delta
                Delta.sim <- matrix(c(D.sim.aux1,D.sim.aux2),ncol=(n-m),byrow=TRUE)
            }
            if(esq.pert=="dispersion")
            {   D.sim.aux1 <- (1/varphi.hat)*diag(C.sim.sim)%*%uwt       # varphi
                D.sim.aux2 <- (2/varphi.hat)*diag(B.sim)%*%O.sim          # delta
                Delta.sim <- matrix(c(D.sim.aux1,D.sim.aux2),ncol=(n-m),byrow=TRUE)
            }

            if(is.null(fixed)=="FALSE")
            {   Delta.sim.new = matrix(0,ncol=(n-m),nrow=(1+table(is.na(fixed))["TRUE"]))
                Delta.sim.new[1,] = Delta.sim[1,]
                l=2
                for(i in 2:nrow(Delta.sim))
                {  if(sum(Delta.sim[i,])!=0)
                   {  Delta.sim.new[l,] = Delta.sim[i,]
                      l=l+1
                   }
                }
                I.sim.new = matrix(0,nrow=(1+table(is.na(fixed))["TRUE"]),ncol=(np+nq+p+2))
                I.sim.new[1,] = I.sim[1,]
                l=2
                for(i in 2:nrow(I.sim))
                {  if(sum(I.sim[i,])!=0)
                   {  I.sim.new[l,] = I.sim[i,]
                      l=l+1
                   }
                }
                I.sim.new.new = matrix(0,nrow=(1+table(is.na(fixed))["TRUE"]),ncol=(1+table(is.na(fixed))["TRUE"]))
                I.sim.new.new[,1] = I.sim.new[,1]
                l=2
                for(i in 2:ncol(I.sim.new))
                {  if(sum(I.sim.new[,i])!=0)
                   {  I.sim.new.new[,l] = I.sim.new[,i]
                      l=l+1
                   }
                }
                Delta.sim = Delta.sim.new
                I.sim = I.sim.new.new
            }

            Curvatura.sim <- -2*t(Delta.sim)%*%solve(I.sim)%*%Delta.sim
            Oc.sim[j] <- max(abs(eigen(Curvatura.sim)$values))
            Curvatura.sim.lv[,j] <- diag(Curvatura.sim)

            if(j==1) print("Benchmarks - 00% ...")   
            if(j==floor(1*rep/4)) print("Benchmarks - 25% ...")
            if(j==floor(2*rep/4)) print("Benchmarks - 50% ...")
            if(j==floor(2*rep/4)) print("Benchmarks - 75% ...")
            if(j==floor(4*rep/4)) print("Benchmarks - 100%")
        }         
        loop = j
        if(loop==rep)  break
      }

      # "BC0" statistic

      BC0 = quantile(Oc.sim,  probs = c(perc.ind1)/100)

      # "BC1" statistic

      K = matrix(nrow = (n-m), ncol = rep)
      for(i in 1:rep)    K[,i] = Curvatura.sim.lv[,i]
      qnt.sim = NULL
      for(i in 1:rep)   qnt.sim[i] = max(abs(K[,i]))
      BC1 = quantile(qnt.sim, probs = c(perc.ind1)/100)
 
      # "BC2" statistic

      bs2 = NULL
      j=1
      for(i in 1:rep)
      {   if(Oc.sim[i] > BC0)
          {  bs2[j] = i
             j = j+1
          }
      }
      help = NULL
      for(i in 1:length(bs2))  
      {   kk = bs2[i]
          help[i] = max(abs(K[,kk]))
      }
      BC2 = quantile(help,  probs = c(perc.ind2)/100)

      # Plot
 
      if(plot=="TRUE")
      {  dat = abs(Curvatura.lv)
         barplot(dat,space=1,xlab="Index",ylab="",main=substitute(paste("Lesaffre & Verbeke")))
         abline(h=BC1,lty=1)
         abline(h=BC2,lty=2)
      }

      # Print 

      if(trace == "TRUE")
      {  
   	   print(matrix(nrow=2, ncol=0, byrow=TRUE,dimnames = list(c("Call:", paste(c("symarma(",np,",",nq,") - family: ",dist),collapse="")))))    
         print(matrix(nrow=1, ncol=0, byrow=TRUE,dimnames = list(c("Lesaffre & Verbeke"))))
         print(matrix(nrow=1, ncol=0, byrow=TRUE,dimnames = list(c("Measures of local influence:"))))
         print(matrix(c(OC.exe),nrow=1, ncol=1,byrow=TRUE,dimnames = list(c("Curvature"))))
         print(matrix(nrow=1, ncol=0, byrow=TRUE,dimnames = list(c("Benchmarks:"))))
         print(matrix(c(BC0,BC1,BC2),nrow=3, ncol=1,byrow=FALSE,dimnames = list(c(
            "BC0","BC1","BC2"))))
      }
   }
}
